from . import load_dataset
from .query import XPathQuery
from .schema import NS_PREFIX


class BIDSLayout:
    def __init__(self, ds_dir: str):
        self.dataset = load_dataset(ds_dir)
        self.sc = self.dataset._schema
        self.query = XPathQuery(self.dataset, self.sc)

    def _query(self, expr: str, search_node=None):
        qry_result = self.query.execute(expr, search_node)
        return qry_result

    def _query_entities(self, entity_key):
        entities = self._query('//bids:entities[@key = "%s"]/@value' % entity_key)
        entities = sorted(list(set(entities)))
        return entities

    def get_subjects(self):
        return self._query_entities('sub')

    def get_sessions(self):
        return self._query_entities('ses')

    def get_tasks(self):
        return self._query_entities('task')

    def _gen_scalar_expr(self, k, v):
        if v is None:
            return 'not(@%s)' % k
        return '@%s="%s"' % (k, v)

    def _scalar_or_list(self, attr_name, v):
        if isinstance(v, list):
            values = list(map(lambda val: self._gen_scalar_expr(attr_name, val), v))
            return '(' + ' or '.join(values) + ')'
        else:
            return self._gen_scalar_expr(attr_name, v)

    def get(self, return_type='object', target=None, scope: str = None, extension=None, suffix=None,
            regex_search=False, absolute_paths=None, invalid_filters='error',
            **entities):
        expr = []
        if scope:
            # TODO split into paths and set the last path as search context
            expr.append('//%s:%s' % (NS_PREFIX, scope))
        entity_filters = []
        for k, v in entities.items():
            v = self.sc.process_entity_value(k, v)
            v = self._scalar_or_list('value', v)
            entity_filters.append('%s:entities[@key="%s" and %s]' % (NS_PREFIX, k, v))
        if extension:
            v = self._scalar_or_list('extension', extension)
            entity_filters.append(v)
        if suffix:
            v = self._scalar_or_list('suffix', suffix)
            entity_filters.append(v)
        if entity_filters:
            entity_filters_str = ' and '.join(entity_filters)
            expr.append('//*[%s]' % entity_filters_str)
        expr_final = ''.join(expr)
        artifacts = self._query(expr_final)
        if return_type and return_type.startswith("file"):
            artifacts = list(map(lambda e: e.get_absolute_path(), artifacts))
        return artifacts



import logging

from .schema import Schema
from .model import Dataset
from . import rules, XPathQuery

logger = logging.getLogger(__file__)


class Validator:
    def __init__(self):
        self.ruleAcceptor = None

    def validate(self, schema: Schema, dataset: Dataset):
        validation_rules = rules.collect_rules()
        report = ValidationReport()
        query = XPathQuery(dataset, schema)
        for rule in validation_rules:
            # if rule is disabled, skip it
            if self.ruleAcceptor is not None and not self.ruleAcceptor(rule):
                continue
            try:
                instance = rule()
                instance.validate(schema=schema, dataset=dataset, report=report, query=query)
            except Exception as e:
                logger.error("Could not execute validation rule %s: %s" % (rule.__name__, e))
                raise e
        return report


class ValidationReport:
    def __init__(self):
        self.messages = []

    def error(self, message):
        self.messages.append({
            'severity': 'error',
            'message': message
        })


class ValidationRule:
    def validate(self, **kwargs):
        raise NotImplemented()



import logging
import os
from collections import OrderedDict

from . import files

logger = logging.getLogger(__file__)

SCHEMA_PATH = os.path.dirname(__file__) + '/data/schema-files'

NS = 'https://bids.neuroimaging.io/1.6'
NS_PREFIX = 'bids'
NS_MAP = {NS_PREFIX: NS}


class Schema:
    def __init__(self):
        self.modalities = files.load_contents(SCHEMA_PATH + "/modalities.yaml")
        self.entities = files.load_contents(SCHEMA_PATH + "/entities.yaml")
        # convert to dictionary for faster lookup by entity keys
        # and keep order of entities in sync with entries in file
        self.entities = OrderedDict([(e['key'], e) for e in self.entities])
        self.datatypes = self.merge_to_dict(SCHEMA_PATH + "/datatypes")

    def merge_to_dict(self, dir_path):
        result = {}
        for file in files.get_files(dir_path, include_folders=False, fnmatch_pattern="*.yaml"):
            name_wo_ext = os.path.basename(file)[:-5]
            result[name_wo_ext] = files.load_contents(file)
        return result

    def _trim_int(self, value):
        try:
            # remove paddings/fillers in index values: 001 -> 1, 000230 -> 230
            return str(int(value))
        except ValueError:
            return value

    def process_entity_value(self, key, value):
        if not value or key not in self.entities:
            return value
        sc_entity = self.entities[key]
        if value and sc_entity and 'format' in sc_entity and sc_entity['format'] == 'index':
            if isinstance(value, list):
                return list(map(lambda v: self._trim_int(v), value))
            else:
                return self._trim_int(value)
        return value


import lxml

from . import model
from . import schema


class Query:
    def __init__(self, dataset: model.Dataset, scm: schema.Schema):
        self.dataset = dataset
        self.scm = scm
        self.id2x = {}
        self.x2id = {}
        self.root = dataset.to_etree(mapping_=self.id2x, reverse_mapping_=self.x2id, nsmap_=schema.NS_MAP)

    def execute(self, expr, search_node: lxml.etree.Element = None):
        raise NotImplemented()


class XPathQuery(Query):
    def __init__(self, dataset: model.Dataset, scm: schema.Schema):
        super(XPathQuery, self).__init__(dataset, scm)

    def execute(self, expr, search_node: lxml.etree.Element = None):
        context = self.root
        if search_node:
            context = search_node
        result = context.xpath(expr, namespaces=schema.NS_MAP)
        result = list(map(lambda e: self.x2id[e] if e in self.x2id else e, result))
        return result


class CSSQuery(Query):
    def __init__(self, dataset: model.Dataset, scm: schema.Schema):
        super(CSSQuery, self).__init__(dataset, scm)

    def execute(self, expr, search_node: lxml.etree.Element = None):
        raise NotImplemented()


class ObjectPathQuery(Query):
    def __init__(self, dataset: model.Dataset, scm: schema.Schema):
        super(ObjectPathQuery, self).__init__(dataset, scm)

    def execute(self, expr, search_node: lxml.etree.Element = None):
        raise NotImplemented()




# This file was generated by 'versioneer.py' (0.20) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2021-08-22T23:29:49+0200",
 "dirty": true,
 "error": null,
 "full-revisionid": "f0b0247e6ce03ace5533fc14a75e3301578d5539",
 "version": "0.0.7"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import os

from . import model
from . import files
from .dsloader import DatasetLoader
from .dssaver import DatasetSaver
from .schema import Schema, NS_MAP
from . import utils


def load_dataset(base_dir: str):
    schema = Schema()
    loader = DatasetLoader(schema)
    ds = loader.load(base_dir)
    return ds


def save_dataset(ds: model.Dataset, target_dir: str):
    saver = DatasetSaver(ds._schema)
    return saver.save(ds, target_dir)


def to_etree(ds: model.Dataset):
    return ds.to_etree(nsmap_=NS_MAP)


# start monkey-patching generated code
from .query import XPathQuery


def load_file_contents(folder: model.Folder, file_name):
    file_path = get_absolute_path(folder, file_name)
    contents = files.load_contents(file_path)
    return contents


setattr(model.Folder, 'load_file_contents', load_file_contents)


def load_contents(file: model.File):
    file_path = get_absolute_path(file.parent_object_, file.name)
    contents = files.load_contents(file_path)
    return contents


setattr(model.File, 'load_contents', load_contents)


def get_absolute_path_by_file(file: model.File):
    return get_absolute_path(file.parent_object_, file.name)


setattr(model.File, 'get_absolute_path', get_absolute_path_by_file)


def get_absolute_path(folder: model.Folder, file_name):
    return _get_path(folder, file_name, True)


def _folder_get_relative_path(folder: model.Folder):
    return _get_path(folder, None, False)


setattr(model.Folder, 'get_relative_path', _folder_get_relative_path)


def _file_get_relative_path(file: model.File):
    return _get_path(file.parent_object_, file.name, False)


setattr(model.File, 'get_relative_path', _file_get_relative_path)


def _get_path(folder: model.Folder, file_name, absolute=True):
    segments = []
    if file_name:
        segments.append(file_name)
    current_folder = folder
    while current_folder is not None:
        if isinstance(current_folder, model.Dataset):
            if absolute:
                segments.insert(0, current_folder.base_dir_)
            # assume we reached the highest level, maybe not good for nested datasets
            break
        else:
            segments.insert(0, current_folder.get_name())
        current_folder = current_folder.parent_object_
    _path = os.path.join(*segments) if segments else ''
    return _path


def remove_file(folder: model.Folder, file_name):
    folder.files = list(filter(lambda file: file.name != file_name, folder.files))


setattr(model.Folder, 'remove_file', remove_file)


def get_file(folder: model.Folder, file_name):
    return next(filter(lambda file: file.name == file_name, folder.files), None)


setattr(model.Folder, 'get_file', get_file)


def remove_folder(folder: model.Folder, folder_name):
    folder.folders = list(filter(lambda f: f.name != folder_name, folder.folders))


setattr(model.Folder, 'remove_folder', remove_folder)


def get_folder(folder: model.Folder, folder_name):
    return next(filter(lambda f: f.name == folder_name, folder.folders), None)


setattr(model.Folder, 'get_folder', get_folder)


def get_files_sorted(folder: model.Folder):
    return sorted(folder.get_files(), key=lambda f: f.name)


setattr(model.Folder, 'get_files_sorted', get_files_sorted)


def get_folders_sorted(folder: model.Folder):
    return sorted(folder.get_folders(), key=lambda f: f.name)


setattr(model.Folder, 'get_folders_sorted', get_folders_sorted)

# end monkey-patching


from .pybids_compat import BIDSLayout

from . import _version

__version__ = _version.get_versions()['version']



import inspect
import os
import shutil

from ancpbids import model
from ancpbids.schema import Schema


class DatasetSaver:
    def __init__(self, schema: Schema):
        pass

    def save(self, ds: model.Dataset, target_dir: str):
        if os.path.exists(target_dir) and len(os.listdir(target_dir)) > 0:
            raise ValueError("Directory not empty: " + target_dir)
        src_dir = ds.base_dir_
        generator = ds.generateRecursively_()
        for obj, _ in generator:
            typ = type(obj)
            mapper_name = '_type_handler_' + typ.__name__
            if mapper_name not in _TYPE_MAPPERS:
                mapper_name = '_type_handler_default'
            mapper = _TYPE_MAPPERS[mapper_name]
            mapper(self, src_dir, target_dir, obj)
        # copy internal children (files/folders)
        self._type_handler_Folder(src_dir, target_dir, ds, traverse_children=True)
        return target_dir

    def _type_handler_default(self, src_dir, target_dir, obj):
        if isinstance(obj, model.Folder):
            self._type_handler_Folder(src_dir, target_dir, obj)
        elif isinstance(obj, model.File):
            self._type_handler_File(src_dir, target_dir, obj)

    def _type_handler_File(self, src_dir, target_dir, file: model.File, new_file_name=None):
        old_file_name = file.get_relative_path()
        old_file_path = os.path.join(src_dir, old_file_name)
        new_file_path = new_file_name
        if new_file_path:
            new_file_path = os.path.join(target_dir, file.parent_object_.get_relative_path(), new_file_path)
        else:
            new_file_path = os.path.join(target_dir, old_file_name)
        shutil.copy(old_file_path, new_file_path)

    def _type_handler_Folder(self, src_dir, target_dir, folder: model.Folder, traverse_children=False):
        new_dir = os.path.join(target_dir, folder.get_relative_path())
        # the new directory may exist because model Artifacts/Folders are processed first
        if not os.path.exists(new_dir):
            os.mkdir(new_dir)

        if traverse_children:
            for child_folder in folder.folders:
                self._type_handler_Folder(src_dir, target_dir, child_folder)
            for child_file in folder.files:
                self._type_handler_File(src_dir, target_dir, child_file)

    def _type_handler_Artifact(self, src_dir, target_dir, artifact: model.Artifact):
        segments = []
        # TODO sort according order defined in schema
        for e in artifact.get_entities():
            seg = '-'.join([e.get_key(), e.get_value()])
            segments.append(seg)
        segments.append(artifact.get_suffix())
        new_file_name = '_'.join(segments) + artifact.get_extension()
        self._type_handler_File(src_dir, target_dir, artifact, new_file_name)


_TYPE_MAPPERS = {name: obj for name, obj in inspect.getmembers(DatasetSaver) if
                 inspect.isfunction(obj) and obj.__name__.startswith('_type_handler_')}



import inspect
import os

import regex

from ancpbids import model
from ancpbids.schema import Schema, NS_PREFIX
from . import utils

ENTITIES_PATTERN = regex.compile(r'(([^\W_]+)-([^\W_]+)_)+([^\W_]+)(.*)')


class DatasetLoader:
    def __init__(self, schema: Schema):
        self.schema = schema

    def load(self, base_dir):
        ds = model.Dataset()
        ds.set_ns_prefix_(NS_PREFIX)
        ds._schema = self.schema
        ds.set_name(os.path.basename(base_dir))
        ds.set_name(os.path.basename(base_dir))
        ds.base_dir_ = base_dir
        # 1. pass: load file system structure
        self._load_folder(ds, base_dir)
        # 2. pass: transform files to artifacts, i.e. files containing entities in their name
        self._convert_files_to_artifacts(ds)
        # 3. pass: expand structure based on schema-files
        self._expand_members(ds)
        return ds

    def _convert_files_to_artifacts(self, parent: model.Folder):
        for i, file in enumerate(parent.files):
            artifact = self._convert_to_artifact(file)
            if not artifact:
                continue
            artifact.parent_object_ = parent
            parent.replace_files_at(i, artifact)
        for folder in parent.folders:
            self._convert_files_to_artifacts(folder)

    def _convert_to_artifact(self, file: model.File):
        match = ENTITIES_PATTERN.match(file.name)
        if not match:
            return None
        artifact = model.Artifact()
        artifact.name = file.name
        for pair in zip(match.captures(2), match.captures(3)):
            entity = model.EntityRef()
            key = pair[0]
            entity.set_key(key)
            value = self.schema.process_entity_value(key, pair[1])
            entity.set_value(value)
            artifact.add_entities(entity)
        artifact.set_suffix(match[4])
        artifact.set_extension(match[5])
        return artifact

    def _get_schema(self, context):
        return self.schema

    def _handle_direct_folders(self, parent, member, pattern, new_type):
        if not isinstance(parent, model.Folder):
            return
        folders = list(filter(lambda f: regex.match(pattern, f.name), parent.get_folders_sorted()))
        for folder in folders:
            obj = new_type()
            obj.name = folder.name
            obj.files = folder.files
            obj.folders = folder.folders
            parent.remove_folder(folder.name)
            obj.parent_object_ = parent
            getattr(parent, member['name']).append(obj)
            self._expand_members(obj)

    def _expand_member(self, parent, member):
        typ = member['typ']
        mapper_name = '_type_handler_' + typ.__name__
        if mapper_name not in _TYPE_MAPPERS:
            mapper_name = '_type_handler_default'
        mapper = _TYPE_MAPPERS[mapper_name]
        mapper(self, parent, member)

    def _expand_members(self, folder: model.Folder):
        members = utils.get_members(type(folder))
        for member in members:
            self._expand_member(folder, member)

    def _load_folder(self, parent: model.Folder, dir_path):
        for root, directories, files in os.walk(dir_path):
            for directory in sorted(directories):
                folder = model.Folder(parent_object_=parent)
                folder.set_name(directory)
                parent.add_folders(folder)
                self._load_folder(folder, root + "/" + directory)
            for file in sorted(files):
                model_file = model.File(parent_object_=parent)
                model_file.set_name(file)
                parent.add_files(model_file)
            break

    def _type_handler_default(self, parent, member):
        typ = member['typ']
        if issubclass(typ, model.JsonFile):
            self._type_handler_JsonFile(parent, member)

    def _type_handler_File(self, parent, member):
        if not isinstance(parent, model.Folder):
            return
        file = parent.get_file(member['name_raw'])
        if file:
            setattr(parent, member['name'], file)
            parent.remove_file(file.name)

    def _type_handler_Artifact(self, parent, member):
        if not isinstance(parent, model.Folder):
            return
        attr = getattr(parent, member['name'])
        multi = isinstance(attr, list)
        name = member['name_raw']
        files = parent.get_files() if multi else list(filter(lambda f: f.name == name, parent.get_files()))
        for file in files:
            if not isinstance(file, model.Artifact):
                continue
            file.parent_object_ = parent
            parent.remove_file(file.name)
            if multi:
                attr.append(file)
            else:
                setattr(parent, member['name'], file)

    def _type_handler_Folder(self, parent, member):
        if not isinstance(parent, model.Folder):
            return
        name = member['name']
        folder = parent.get_folder(name)
        if folder:
            setattr(parent, name, folder)
            parent.remove_folder(name)

    def _type_handler_Subject(self, parent, member):
        self._handle_direct_folders(parent, member, "sub-", model.Subject)

    def _type_handler_Session(self, parent, member):
        self._handle_direct_folders(parent, member, "ses-", model.Session)

    def _type_handler_DatatypeFolder(self, parent, member):
        pattern = '|'.join(self.schema.datatypes.keys())
        self._handle_direct_folders(parent, member, pattern, model.DatatypeFolder)

    def _type_handler_JsonFile(self, parent, member):
        name = member['name_raw']
        json_object = parent.load_file_contents(name)
        if not json_object:
            return
        model_type = member['typ']
        dsd_file = model_type()
        dsd_file.name = name
        members = utils.get_members(model_type, False)
        actual_props = json_object.keys()
        direct_props = list(map(lambda m: m['name'], members))
        for prop_name in direct_props:
            if prop_name in actual_props:
                value = json_object[prop_name]
                setattr(dsd_file, prop_name, value)
        setattr(parent, member['name'], dsd_file)
        parent.remove_file(name)


_TYPE_MAPPERS = {name: obj for name, obj in inspect.getmembers(DatasetLoader) if
                 inspect.isfunction(obj) and obj.__name__.startswith('_type_handler_')}



from ancpbids import XPathQuery
from ancpbids.model import Dataset, GdsCollector_, Session, Subject, Datatype
from ancpbids.schema import Schema
from ancpbids.validator import ValidationRule, ValidationReport


class StaticStructureValidationRule(ValidationRule):
    def validate(self, dataset: Dataset, report: ValidationReport, **kwargs):
        collector = GdsCollector_()
        dataset.validate_(collector, recursive=True)
        for message in collector.messages:
            report.error(message)


class DatatypesValidationRule(ValidationRule):
    def validate(self, dataset: Dataset, report: ValidationReport, **kwargs):
        dangling_folders = []
        for subject in dataset.get_subjects():
            dangling_folders.extend(subject.get_folders())
            for session in subject.get_sessions():
                dangling_folders.extend(session.get_folders())

        for folder in dangling_folders:
            report.error("Unsupported datatype folder '%s'" % folder.get_relative_path())


class EntitiesValidationRule(ValidationRule):
    def validate(self, schema: Schema, dataset: Dataset, report: ValidationReport, query: XPathQuery):
        artifacts = query.execute('//bids:entities/..')
        expected_key_order = {k: i for i, k in enumerate(schema.entities.keys())}
        expected_order_key = {i: k for i, k in enumerate(schema.entities.keys())}
        for artifact in artifacts:
            entity_refs = artifact.get_entities()
            found_invalid_key = False
            for ref in entity_refs:
                if ref.key not in schema.entities:
                    report.error(
                        "Invalid entity '%s' in artifact '%s'" % (ref.key, artifact.get_relative_path()))
                    found_invalid_key = True
            if found_invalid_key:
                # we cannot check the order of entities if invalid entity found
                continue
            # now, check if order of entities matches order in schema
            keys = list(map(lambda e: e.key, entity_refs))
            actual_keys_order = list(map(lambda k: expected_key_order[k], keys))
            for i in range(0, len(actual_keys_order) - 1):
                if actual_keys_order[i] > actual_keys_order[i + 1]:
                    expected = tuple(map(lambda k: expected_order_key[k], sorted(actual_keys_order)))
                    report.error(
                        "Invalid entities order: expected=%s, found=%s, artifact=%s" % (
                        expected, tuple(keys), artifact.get_relative_path()))
                    break


class SuffixesValidationRule(ValidationRule):
    def validate(self, schema: Schema, dataset: Dataset, report: ValidationReport, query: XPathQuery):
        pass



def collect_rules():
    rules = []

    # use dynamic import to prevent circular dependencies when module is loaded
    import ancpbids.rules.rules_files as sr
    rules.append(sr.StaticStructureValidationRule)
    rules.append(sr.DatatypesValidationRule)
    rules.append(sr.EntitiesValidationRule)

    return rules



import inspect

from ancpbids import model

_MODEL_CLASSES = {name: obj for name, obj in inspect.getmembers(model) if inspect.isclass(obj)}

def get_members(element_type, include_superclass=True):
    if element_type == model.File or element_type == model.Folder:
        return []
    super_members = []

    if include_superclass:
        try:
            if element_type.superclass:
                super_members = get_members(element_type.superclass, include_superclass)
        except AttributeError:
            pass

    element_members = []
    try:
        # name is the class member name compatible with Python naming conventions
        # name_raw contains the name as modeled in schema and may contain invalid characters such as dots
        element_members = list(
            map(lambda member: {'name_raw': member.child_attrs['name'], 'type': member.data_type, **member.child_attrs,
                                'name': member.name},
                element_type.member_data_items_.values()))
        element_members = list(map(_normalize, element_members))
    except AttributeError as ae:
        pass
    return super_members + element_members

def _to_type(model_type_name: str):
    if model_type_name == 'string':
        return str
    return _MODEL_CLASSES[model_type_name]


def _normalize(member):
    result = {'name': member['name'], 'name_raw': member['name_raw'], 'typ': _to_type(member['type'])}
    if 'use' in member:
        result['lower'] = 1 if member['use'] == 'required' else 0
        result['upper'] = 1
    else:
        result['lower'] = member['minOccurs']
        result['upper'] = member['maxOccurs']
    return result
